from unittest.mock import MagicMock, patch

import pytest

from pull_request_codecommit.git import Commits
from pull_request_codecommit.pull_request import PullRequest


COMMIT_SIMPLE_MESSAGE = """commit my-hash-1
Author: John Doe <john@doe.nl>
Date:   Fri Jan 21 21:01:00 2022 +0100

    feat: my first commit"""

COMMIT_DETAILED_MESSAGE = """commit my-hash-1
Author: John Doe <john@doe.nl>
Date:   Fri Jan 21 21:01:00 2022 +0100

    feat: my first commit

    Some additional information"""

COMMIT_SIMPLE_MESSAGE_ISSUE = """commit my-hash-1
Author: John Doe <john@doe.nl>
Date:   Fri Jan 21 21:01:00 2022 +0100

    feat: my first commit

    Issue: #1"""

COMMIT_DETAILED_MESSAGE_ISSUE = """commit my-hash-1
Author: John Doe <john@doe.nl>
Date:   Fri Jan 21 21:01:00 2022 +0100

    feat: my first commit

    Some additional information

    Issue: #1"""


@pytest.mark.parametrize(
    "commits, expected_title, expected_description",
    [
        (Commits(COMMIT_SIMPLE_MESSAGE), "feat: my first commit", ""),
        (
            Commits(COMMIT_DETAILED_MESSAGE),
            "feat: my first commit",
            "Some additional information",
        ),
        (
            Commits(COMMIT_SIMPLE_MESSAGE_ISSUE),
            "feat: my first commit (#1)",
            "Issues: #1",
        ),
        (
            Commits(COMMIT_DETAILED_MESSAGE_ISSUE),
            "feat: my first commit (#1)",
            "Some additional information\n\nIssues: #1",
        ),
        (
            Commits(""),
            "",
            "",
        ),
    ],
)
@patch("pull_request_codecommit.pull_request_codecommit.AwsClient")
def test_git_client(
    mock_aws_client: MagicMock,
    commits: Commits,
    expected_title: str,
    expected_description: str,
) -> None:
    mock_aws_client.return_value.get_open_pull_request.return_value = {}

    mock_repo = MagicMock()
    mock_repo.commits.return_value = commits

    pull_request = PullRequest(mock_repo)
    assert pull_request.title == expected_title
    assert pull_request.description == expected_description


@patch("pull_request_codecommit.pull_request_codecommit.AwsClient")
def test_update_pull_request(mock_aws_client: MagicMock) -> None:
    mock_aws_client.return_value.get_open_pull_request.return_value = {
        "pullRequestId": "1",
        "title": "feat: test update pull request",
        "authorArn": "arn:aws:sts::111122223333:assumed-role/Role/john.doe@acme.com",
        "description": "my committed description",
    }

    mock_repo = MagicMock()
    mock_repo.commits.return_value = Commits(COMMIT_DETAILED_MESSAGE_ISSUE)
    mock_repo.remote.name = "my-repository"
    mock_repo.remote.region = "eu-west-1"
    pull_request = PullRequest(mock_repo)

    assert pull_request.title == "feat: test update pull request"
    assert "my committed description" in pull_request.description
    assert "Existing pull request found: #1" in pull_request.description
    assert "Author: john.doe@acme.com" in pull_request.description
    assert "Issues: #1" in pull_request.description

    pull_request.proposal.update(pull_request.title, pull_request.description)
    pull_request.save()



COMMIT2 = """commit my-hash-1
Author: John Doe <john@doe.nl>
Date:   Fri Jan 21 21:01:00 2022 +0100

    feat: my first commit"""

COMMIT1 = """commit my-hash-2
Author: John Doe <john@doe.nl>
Date:   Fri Jan 21 21:01:00 2022 +0100

    feat: my second commit"""

COMMITS = f"{COMMIT1}\n\n    Issue #1\n\n{COMMIT2}\n\n    Issue #2"
COMMITS_NO_ISSUES = f"{COMMIT1}\n\n{COMMIT2}"

SCENARIOS = [
    (
        "codecommit::eu-west-1://my-profile@my-repository",
        "eu-west-1",
        "my-profile",
        b"[default]\nbranch: my-main\n[profile my-profile]\nbranch: my-master",
        COMMITS,
    ),
    (
        "codecommit::eu-west-1://my-profile@my-repository",
        "eu-central-1",
        "my-profile",
        b"[default]\nbranch: my-main\n[profile my-profile]\nbranch: my-master",
        COMMITS_NO_ISSUES,
    ),
    (
        "codecommit::eu-west-1://my-profile@my-repository",
        "eu-west-1",
        "my-other-profile",
        b"[default]\nbranch: my-main\n[profile my-profile]\nbranch: my-master",
        COMMITS_NO_ISSUES,
    ),
    (
        "codecommit::eu-west-1://my-profile@my-repository",
        "eu-central-1",
        "my-other-profile",
        b"[default]\nbranch: my-main\n[profile my-profile]\nbranch: my-master",
        COMMITS,
    ),
    (
        "codecommit::eu-west-1://my-repository",
        "eu-central-1",
        None,
        b"[default]\nbranch: my-main\n[profile my-profile]\nbranch: my-master",
        COMMITS,
    ),
    (
        "codecommit::://my-profile@my-repository",
        None,
        "my-profile",
        b"[default]\nbranch: my-main\n[profile my-profile]\nbranch: my-master",
        COMMITS,
    ),
    (
        "codecommit::://my-repository",
        None,
        None,
        b"[default]\nbranch: my-main\n[profile my-profile]\nbranch: my-master",
        COMMITS,
    ),
    (
        "codecommit://my-profile@my-repository",
        None,
        "my-profile",
        b"[default]\nbranch: my-main\n[profile my-profile]\nbranch: my-master",
        COMMITS,
    ),
    (
        "codecommit://my-repository",
        None,
        None,
        b"[default]\nbranch: my-main\n[profile my-profile]\nbranch: my-master",
        COMMITS,
    ),
    (
        "codecommit://my-repository-pr-failure",
        None,
        None,
        b"[default]\nbranch: my-main\n[profile my-profile]\nbranch: my-master",
        COMMITS,
    ),
    (
        "codecommit://my-repository-open-pr",
        None,
        None,
        b"[default]\nbranch: my-main\n[profile my-profile]\nbranch: my-master",
        COMMITS,
    ),
    (
        "codecommit://my-repository-other-open-pr",
        None,
        None,
        b"[default]\nbranch: my-main\n[profile my-profile]\nbranch: my-master",
        COMMITS,
    ),
]



from typing import List, Optional
import re

from .commit import Commit
from .message import Message


class Commits:
    """
    Understands git commits
    """

    __index: int

    def __init__(self, messages: str) -> None:
        self.__index = 0
        self.__commits: List[Commit] = list(
            map(
                self.__read_commit_message,
                map(
                    lambda x: f"\ncommit {x}",
                    list(filter(None, messages.split("\ncommit "))),
                ),
            )
        )
        self.__commits.reverse()

    @property
    def first(self) -> Optional[Commit]:
        return self.__commits[0] if self.__commits else None

    def __iter__(self):
        """
        We will skip the first commit that can be retrieved by using the `first` property.
        """
        return iter(self.__commits[1:])

    def __next__(self) -> Commit:
        item = self.__commits[self.__index]
        self.__index += 1
        return item

    def __len__(self) -> int:
        return len(self.__commits)

    @property
    def issues(self) -> List[str]:
        return list(
            set(filter(None, map(lambda commit: commit.message.issue, self.__commits)))
        )

    @staticmethod
    def __read_commit_message(message: str) -> Commit:
        def extract(regex: str) -> str:
            match = re.search(regex, message)
            return match.group(1).strip(" ") if match else ""

        return Commit(
            commit=extract(r"^commit (.*)"),
            author=extract(r"Author: (.*)"),
            date=extract(r"Date: (.*)"),
            message=Message(message=extract(r"(?sm)^    (.*)")),
        )



import os
from typing import Optional

from .config import Config
from .git import Client as GitClient, Remote, Commits


class Repository:
    """
    Understands CodeCommit repositories
    """

    def __init__(
        self, path: Optional[str] = None, target_branch: Optional[str] = None
    ) -> None:
        if not path:
            path = os.getcwd()

        self.__git = GitClient(path)
        self.__remote: Optional[Remote] = None
        self.__branch: str = ""
        self.__target_branch: str = (
            target_branch if target_branch else self.__config("destination_branch")
        )

    def __config(self, method: str) -> str:
        item = getattr(Config, method)(self.remote.profile)
        return item if item else ""

    @property
    def remote(self) -> Remote:
        if not self.__remote:
            self.__remote = Remote(self.__git.remote("origin"))

        return self.__remote

    @property
    def branch(self) -> str:
        if not self.__branch:
            self.__branch = self.__git.current_branch

        return self.__branch

    @property
    def destination(self) -> str:
        return self.__target_branch

    def commits(self) -> Commits:
        return self.__git.get_commit_messages(destination_branch=self.destination)

    def push(self) -> None:
        self.__git.push()

    def checkout_destination(self) -> None:
        branch = self.branch
        self.__git.checkout(self.destination)
        self.__git.pull()
        self.__git.delete_branch(branch)



from typing import Dict, Optional

import configparser
from pathlib import Path


class Config:
    """
    Understands the configuration of the pull-request-codecommit tool.
    """

    __cached_config: Dict[str, Dict[str, str]] = {}
    __config_file: Optional[configparser.ConfigParser] = None

    @classmethod
    def __default_config(cls) -> Dict[str, str]:
        return cls.__load_config("default")

    @classmethod
    def __get_config_value(cls, profile: str, key: str) -> str:
        if profile not in cls.__cached_config:
            cls.__cached_config[profile] = cls.__default_config()
            cls.__cached_config[profile].update(cls.__load_config(profile))

        return cls.__cached_config[profile].get(key, "")

    @classmethod
    def __user_config(cls) -> configparser.ConfigParser:
        if not cls.__config_file:
            cls.__config_file = configparser.ConfigParser()
            cls.__config_file.read(f"{Path.home()}/.aws/pull-request-codecommit")

        return cls.__config_file

    @classmethod
    def __load_config(cls, profile: str) -> Dict[str, str]:
        config = {}

        if profile != "default":
            profile = f"profile {profile}"

        if profile in cls.__user_config().sections():
            for key, value in list(cls.__user_config().items(profile)):
                config[key] = value

        return config

    @classmethod
    def destination_branch(cls, profile: str) -> str:
        return cls.__get_config_value(profile, "branch")



import re
from typing import Optional


class Remote:
    """
    Understands remote CodeCommit URLs
    """

    def __init__(self, url: str):
        self.__url: str = url
        self.__region: Optional[str] = ""
        self.__profile: Optional[str] = ""
        self.__name: str = ""

    def __regex(self, pattern: str, index: int = 1) -> Optional[str]:
        match = re.search(pattern, self.__url)
        return match.group(index) if match else None

    @property
    def supported(self) -> bool:
        return self.__url.startswith("codecommit:") and self.name != ""

    @property
    def url(self) -> str:
        return self.__url

    @property
    def region(self) -> Optional[str]:
        if not self.__region:
            self.__region = self.__regex(r"^codecommit::(.*)://")

        return self.__region

    @property
    def profile(self) -> Optional[str]:
        if self.__profile == "":
            self.__profile = self.__regex(r"//(.*)@")

        return self.__profile

    @property
    def name(self) -> str:
        if not self.__name:
            name = self.__regex(r"(\/\/|.*@)(.*)$", 2)

            if name:
                self.__name = name

        return self.__name

    @property
    def codecommit_url(self) -> str:
        return f"https://{self.region}.console.aws.amazon.com/codesuite/codecommit/repositories/{self.name}"

    def pull_request_url(self, pull_request_id: int) -> str:
        return (
            self.codecommit_url
            + f"/pull-requests/{pull_request_id}/details?region={self.region}"
        )



from typing import Tuple, Optional

from .repository import Repository
from .aws import Client as AwsClient


class PullRequestCodeCommit:
    """
    Understands CodeCommit Pull requests
    """

    def __init__(self, repository: Repository) -> None:
        self.__aws_client: Optional[AwsClient] = None
        self.__repo = repository
        self.__raw_data: dict = {}

    @property
    def __data(self) -> dict:
        if not self.__raw_data:
            self.__raw_data = self.__client.get_open_pull_request(
                repository=self.__repo.remote.name,
                source=self.__repo.branch,
                destination=self.__repo.destination,
            )

        return self.__raw_data

    @property
    def __client(self) -> AwsClient:
        if not self.__aws_client:
            self.__aws_client = AwsClient(
                profile=self.__repo.remote.profile, region=self.__repo.remote.region
            )

        return self.__aws_client

    @property
    def pull_request_id(self) -> int:
        return self.__data.get("pullRequestId", 0)

    @property
    def author(self) -> str:
        return self.__data.get("authorArn", "").split("/")[-1]

    @property
    def title(self) -> str:
        return self.__data.get("title", "")

    @property
    def description(self) -> str:
        return self.__data.get("description", "").replace("\n\n", "\n")

    @property
    def link(self) -> str:
        return self.__repo.remote.pull_request_url(self.pull_request_id)

    def save(self, title: str, description: str):
        self.__update(
            description=self.__markdown_conversion(description)
        ) if self.pull_request_id != 0 else self.__create(
            title=title, description=self.__markdown_conversion(description)
        )

    def merge(self) -> str:
        response = self.__client.merge_pull_request(
            repository=self.__repo.remote.name,
            pull_request_id=self.pull_request_id,
            branch=self.__repo.branch,
        )
        status = response.get("pullRequestStatus", "")

        if status == "CLOSED":
            self.__repo.checkout_destination()

        return status

    @staticmethod
    def __markdown_conversion(description: str) -> str:
        return description.replace("\n", "\n\n")

    def __create(self, title: str, description: str) -> None:
        self.__raw_data = self.__client.create_pull_request(
            title=title,
            description=description,
            repository=self.__repo.remote.name,
            source=self.__repo.branch,
            destination=self.__repo.destination,
        )

    def __update(self, description: str) -> None:
        self.__client.update_pull_request(
            pull_request_id=self.pull_request_id,
            description=description,
        )



from typing import Optional

import click

from .pull_request import PullRequest
from .repository import Repository


@click.command()
@click.option("-r", "--repository-path", default=None)
@click.option("-b", "--branch", default=None)
@click.option("--auto-merge/--no-auto-merge", default=False)
def main(
    repository_path: Optional[str], branch: Optional[str], auto_merge: bool
) -> None:
    """
    pull-request-codecommit
    """
    repo = __load_repository(repository_path=repository_path, target_branch=branch)
    __display_repository_information(repo)
    pr = __create_pull_request(repo)

    if auto_merge:
        status = pr.merge()
        click.echo(f"Auto merging resulted in: {status}")


def __load_repository(
    repository_path: Optional[str], target_branch: Optional[str]
) -> Repository:
    repo = Repository(path=repository_path, target_branch=target_branch)

    if not repo.remote.supported:
        raise click.ClickException("The repository is not compatible with this tool!")

    return repo


def __create_pull_request(repo: Repository) -> PullRequest:
    pr = PullRequest(repo)

    if not pr.has_changes:
        click.echo(
            f"There are no differences between {repo.destination} and {repo.branch}!"
        )
        exit(0)

    __propose_title_description(pr)
    pr.save()

    click.echo(f"Please review/approve: {pr.link}")
    click.echo()
    click.echo(click.style(pr.title, bold=True))
    click.echo(pr.description)

    return pr


def __display_repository_information(repo: Repository) -> None:
    click.echo(f"Remote: {repo.remote.url}")
    click.echo(f"Region: {repo.remote.region}")
    click.echo(f"Profile: {repo.remote.profile}")
    click.echo(f"Repo: {repo.remote.name}")
    click.echo(f"Branch: {repo.branch}")
    click.echo()


def __propose_title_description(pr: PullRequest) -> None:
    message = click.edit(f"{pr.title}\n\n{pr.description}")

    if message is None:
        raise click.ClickException("Pull request was not created")

    title = message.splitlines()[0]
    description = "\n".join(message.splitlines()[1:]).lstrip("\n")

    pr.update(title, description)


if __name__ == "__main__":
    main()



import re
from typing import List


class Message:
    """
    Understands git message that follow the conventional commits standard.

    See: https://www.conventionalcommits.org/en/v1.0.0/
    """

    def __init__(self, message: str) -> None:
        lines = []

        for line in iter(message.splitlines()):
            lines.append(line.strip(" "))

        self.__issue: str = ""
        self.__lines: List[str] = list(filter(None, lines))

        if "issue" in self.__lines[-1].lower():
            self.__detect_issue(self.__lines.pop())

    def __detect_issue(self, line: str) -> None:
        match = re.search(r"(?i)issue(:|) (.*)", line)
        self.__issue = f"{match.group(2)}" if match else ""

    @property
    def issue(self) -> str:
        """
        Issue reference should be in the footer of the commit message
        """
        return self.__issue

    @property
    def subject(self) -> str:
        """
        Subject is the first line of the commit message
        """
        return self.__lines[0] if len(self.__lines) > 0 else ""

    @property
    def body(self) -> str:
        """
        Body is all except the first line
        """
        return "\n".join(self.__lines[1:]).lstrip("\n")



from typing import Optional

from .pull_request_codecommit import PullRequestCodeCommit
from .pull_request_proposal import PullRequestProposal
from .repository import Repository
from .git import Commits


class PullRequest:
    """
    Understands pull requests
    """

    def __init__(self, repo: Repository) -> None:
        self.__repo: Repository = repo
        self.__proposal: Optional[PullRequestProposal] = None
        self.__commits: Commits = self.__repo.commits()
        self.__pull_request: PullRequestCodeCommit = PullRequestCodeCommit(
            repository=repo
        )

    @property
    def proposal(self) -> PullRequestProposal:
        if not self.__proposal:
            self.__proposal = PullRequestProposal(self.__pull_request, self.__commits)

        return self.__proposal

    @property
    def title(self) -> str:
        return self.proposal.title

    @property
    def description(self) -> str:
        return self.proposal.description

    @property
    def link(self) -> str:
        return self.__pull_request.link

    @property
    def has_changes(self) -> bool:
        return True if self.__commits.first else False

    def save(self) -> None:
        self.__repo.push()
        self.__pull_request.save(self.title, self.description)

    def update(self, title: str, description: str) -> None:
        self.proposal.update(title=title, description=description)

    def merge(self) -> str:
        return self.__pull_request.merge()



import pytest
from unittest.mock import patch, MagicMock
from pull_request_codecommit.repository import Repository


@pytest.mark.parametrize(
    "region, profile, repository",
    [
        ("eu-west-1", "my-profile", "my-repository"),
        ("eu-central-1", "my-other-profile", "my-other-repository"),
    ],
)
@patch("pull_request_codecommit.repository.GitClient")
def test_remote(
    mock_git_client: MagicMock, region: str, profile: str, repository: str
) -> None:
    mock_git_client.return_value.remote.return_value = (
        f"codecommit::{region}://{profile}@{repository}"
    )
    repo = Repository("/my/repository")
    assert repo.remote.region == region
    assert repo.remote.profile == profile
    assert repo.remote.name == repository



from .message import Message


class Commit:
    """
    Understands git commit
    """

    def __init__(self, commit: str, author: str, date: str, message: Message) -> None:
        self.__commit: str = commit
        self.__author: str = author
        self.__date: str = date
        self.__message: Message = message

    @property
    def message(self) -> Message:
        return self.__message



from __future__ import annotations
from typing import List
import os.path
import subprocess
from .commits import Commits


class Client:
    """
    Understands git operations
    """

    def __init__(self, path: str):
        if not os.path.isdir(path):
            raise NotADirectoryError(f"The {path} is not a directory.")

        self.__path: str = path

    def __execute(self, parameters: List[str]) -> str:
        parameters.insert(0, "git")
        response = subprocess.run(parameters, cwd=self.__path, stdout=subprocess.PIPE)

        return response.stdout.decode("utf-8").strip("\n")

    def remote(self, name: str = "origin") -> str:
        return self.__execute(["config", "--get", f"remote.{name}.url"])

    def push(self) -> None:
        self.__execute(["push", "--set-upstream", "origin", "HEAD"])

    @property
    def current_branch(self) -> str:
        return self.__execute(["branch", "--show-current"])

    def get_commit_messages(self, destination_branch: str) -> Commits:
        self.fetch()
        messages = self.__execute(
            ["log", f"origin/{destination_branch}..origin/{self.current_branch}"]
        )

        return Commits(messages)

    def checkout(self, destination: str) -> None:
        self.__execute(["checkout", destination])

    def fetch(self) -> None:
        self.__execute(["fetch"])

    def pull(self) -> None:
        self.__execute(["pull"])

    def delete_branch(self, branch: str) -> None:
        self.__execute(["branch", "-d", branch])



import json
from io import TextIOWrapper, BytesIO
from typing import List, Tuple, Optional, Dict, Any
from unittest.mock import patch, MagicMock

import pytest
from click.testing import CliRunner
from pull_request_codecommit import main
from pull_request_codecommit.config import configparser
from pull_request_codecommit.git import Commits
from tests import SCENARIOS


def edit_message(message: str) -> str:
    return message


def aws_client_execute_side_effect(parameters, stdout) -> MagicMock:
    data: Dict[str, Any] = {}
    assert -1 == stdout
    mock_stdout = MagicMock()

    if "create-pull-request" in parameters:
        data = {"pullRequest": {"pullRequestId": "1"}}

    elif "merge-pull-request-by-fast-forward" in parameters:
        status = "OPEN" if "my-repository-pr-failure" in parameters else "CLOSED"
        data = {"pullRequest": {"pullRequestStatus": status}}

    elif "list-pull-requests" in parameters:
        repository_name = next(
            filter(lambda item: item.startswith("my-repository"), parameters)
        )
        pull_request_ids: List[int] = {
            "my-repository-open-pr": [1],
            "my-repository-other-open-pr": [2],
        }.get(repository_name, [])

        data = {"pullRequestIds": list(map(str, pull_request_ids))}

    elif "get-pull-request" in parameters:
        pull_request_id = int(parameters[parameters.index("--pull-request-id") + 1])

        source = "feat/my-feature" if pull_request_id == 1 else "feat/my-other-feature"
        destination = "my-main" if pull_request_id == 1 else "my-development"

        data = {
            "pullRequest": {
                "pullRequestId": str(pull_request_id),
                "authorArn": "arn:aws:sts::111122223333:assumed-role/Role/john.doe@acme.com",
                "title": "feat: update existing pull request support",
                "description": "Update the existing pull request when a pull request already exists",
                "pullRequestTargets": [
                    {
                        "sourceCommit": "ca45e279EXAMPLE",
                        "sourceReference": f"refs/heads/{source}",
                        "mergeBase": "a99f5ddbEXAMPLE",
                        "destinationReference": f"refs/heads/{destination}",
                        "mergeMetadata": {"isMerged": "false"},
                        "destinationCommit": "2abfc6beEXAMPLE",
                        "repositoryName": "my-repository-open-pr",
                    }
                ],
            }
        }
    elif "update-pull-request-description" in parameters:
        pull_request_id = int(parameters[parameters.index("--pull-request-id") + 1])
        data = {
            "pullRequest": {
                "pullRequestId": str(pull_request_id),
                "authorArn": "arn:aws:sts::111122223333:assumed-role/Role/john.doe@acme.com",
            }
        }

    mock_stdout.stdout = bytes(json.dumps(data), "utf-8")
    return mock_stdout


def generate_invoke_parameters(
    cli_parameters,
) -> List[Tuple[str, Optional[str], Optional[str], bytes, str, List[str]]]:
    """
    Generate pytest parameters that are used to invoke the test
    """
    pytest_parameters = []

    for cli_parameter in cli_parameters:
        for scenario in SCENARIOS:
            pytest_parameters.append(scenario + (cli_parameter,))

    return pytest_parameters


@pytest.mark.parametrize(
    "remote, region, profile, config, commits, parameters",
    generate_invoke_parameters(
        [
            [],
            ["--auto-merge"],
            ["-b", "my-target-branch"],
            ["--branch", "my-target-branch"],
        ]
    ),
)
@patch("pull_request_codecommit.aws.client.subprocess.run")
@patch("pull_request_codecommit.repository.GitClient")
@patch("pull_request_codecommit.click.edit")
def test_invoke(
    mock_edit: MagicMock,
    mock_git_client: MagicMock,
    mock_aws_client: MagicMock,
    remote: str,
    region: str,
    profile: str,
    config: bytes,
    commits: str,
    parameters: List[str],
) -> None:
    mock_edit.side_effect = edit_message
    mock_git_client.return_value.get_commit_messages.return_value = Commits(commits)

    mock_git_client.return_value.remote.return_value = remote
    mock_git_client.return_value.current_branch = "feat/my-feature"

    configparser.open = MagicMock(return_value=TextIOWrapper(BytesIO(config)))  # type: ignore
    mock_aws_client.side_effect = aws_client_execute_side_effect

    runner = CliRunner()
    result = runner.invoke(main, parameters)
    assert result.exit_code == 0


@patch("pull_request_codecommit.repository.GitClient")
@patch("pull_request_codecommit.click.edit")
def test_invoke_no_commits(
    mock_edit: MagicMock,
    mock_git_client: MagicMock,
) -> None:
    mock_edit.side_effect = edit_message
    mock_git_client.return_value.get_commit_messages.return_value = Commits("")

    mock_git_client.return_value.remote.return_value = (
        f"codecommit::eu-west-1://my-profile@my-repository"
    )
    mock_git_client.return_value.current_branch.return_value = "feat/my-feature"
    config = b"[default]\nbranch: my-main\n[profile my-profile]\nbranch: my-master"
    configparser.open = MagicMock(return_value=TextIOWrapper(BytesIO(config)))  # type: ignore

    runner = CliRunner()
    result = runner.invoke(main, [])
    assert result.exit_code == 0
    assert "There are no differences between" in result.output


@pytest.mark.parametrize(
    "region, profile, config",
    [
        (
            "eu-west-1",
            "my-profile",
            b"[default]\nbranch: my-main\n[profile my-profile]\nbranch: my-master",
        ),
        (
            "eu-central-1",
            "my-profile",
            b"[default]\nbranch: my-main\n[profile my-profile]\nbranch: my-master",
        ),
        (
            "eu-west-1",
            "my-other-profile",
            b"[default]\nbranch: my-main\n[profile my-profile]\nbranch: my-master",
        ),
        (
            "eu-central-1",
            "my-other-profile",
            b"[default]\nbranch: my-main\n[profile my-profile]\nbranch: my-master",
        ),
    ],
)
@patch("pull_request_codecommit.aws.client.subprocess.run")
@patch("pull_request_codecommit.repository.GitClient")
@patch("pull_request_codecommit.click.edit")
def test_invoke_with_path(
    mock_edit: MagicMock,
    mock_git_client: MagicMock,
    mock_aws_client: MagicMock,
    region: str,
    profile: str,
    config: bytes,
) -> None:
    mock_edit.side_effect = edit_message
    mock_git_client.return_value.remote.return_value = (
        f"codecommit::{region}://{profile}@my-repository"
    )
    mock_git_client.return_value.current_branch.return_value = "feat/my-feature"
    configparser.open = MagicMock(return_value=TextIOWrapper(BytesIO(config)))  # type: ignore
    mock_aws_client.side_effect = aws_client_execute_side_effect

    runner = CliRunner()
    result = runner.invoke(main, ["--repository-path", "./some/path/to/repo"])
    assert result.exit_code == 0


@patch("pull_request_codecommit.repository.GitClient")
def test_invoke_github(mock_git_client: MagicMock) -> None:
    mock_git_client.return_value.remote.return_value = (
        "git@github.com:Nr18/pull-request-codecommit.git"
    )
    mock_git_client.return_value.current_branch.return_value = "feat/my-feature"

    runner = CliRunner()
    result = runner.invoke(main, [])
    assert result.exit_code == 1


@patch("pull_request_codecommit.aws.client.subprocess.run")
@patch("pull_request_codecommit.repository.GitClient")
@patch("pull_request_codecommit.click.edit")
def test_invoke_quit_edit(
    mock_edit: MagicMock,
    mock_git_client: MagicMock,
    mock_aws_client: MagicMock,
) -> None:
    mock_edit.return_value = None
    mock_git_client.return_value.remote.return_value = (
        f"codecommit::eu-west-1://my-profile@my-repository"
    )
    mock_git_client.return_value.current_branch.return_value = "feat/my-feature"
    config = b"[default]\nbranch: my-main\n[profile my-profile]\nbranch: my-master"
    configparser.open = MagicMock(return_value=TextIOWrapper(BytesIO(config)))  # type: ignore
    mock_aws_client.side_effect = aws_client_execute_side_effect

    runner = CliRunner()
    result = runner.invoke(main)
    assert type(result.exception) == SystemExit
    assert "Pull request was not created" in result.output
    assert result.exit_code == 1


@patch("pull_request_codecommit.repository.GitClient")
def test_invoke_no_repository_name(
    mock_git_client: MagicMock,
) -> None:
    mock_git_client.return_value.remote.return_value = f"codecommit::eu-west-1://"
    mock_git_client.return_value.current_branch.return_value = "feat/my-feature"
    config = b"[default]\nbranch: my-main\n[profile my-profile]\nbranch: my-master"
    configparser.open = MagicMock(return_value=TextIOWrapper(BytesIO(config)))  # type: ignore

    runner = CliRunner()
    result = runner.invoke(main)
    assert result.exit_code == 1
    assert "Error: The repository is not compatible with this tool!" in result.output



from .git import Commits
from .pull_request_codecommit import PullRequestCodeCommit

SPLITTER = """
#=
#= Existing pull request found: #{{PULL_REQUEST_ID}}
#=
#= Author: {{AUTHOR}}
#=
#= Above you will see the title and description of the existing pull request.
#= Below you will find the newly proposed description.
#=
"""


class PullRequestProposal:
    def __init__(self, pull_request: PullRequestCodeCommit, commits: Commits):
        self.__title = ""
        self.__description = ""
        self.__pull_request = pull_request
        self.__commits = commits

    @property
    def title(self) -> str:
        if self.__title:
            return self.__title

        if self.__pull_request.title:
            return self.__pull_request.title

        return self.__generate_title()

    def __generate_title(self) -> str:
        if not self.__commits.first:
            return ""

        title = self.__commits.first.message.subject

        if self.__commits.issues:
            title += f" ({', '.join(self.__commits.issues)})"

        return title

    @property
    def description(self) -> str:
        if self.__description:
            return self.__description

        if (
            len(self.__commits) == 1
            and self.__commits.first
            and not self.__pull_request.description
        ):
            return self.append_issues(self.__commits.first.message.body)

        return self.__merge_descriptions(
            existing_description=self.__pull_request.description,
            description=self.__generate_description(),
        )

    def append_issues(self, description: str) -> str:
        if not self.__commits.issues:
            return description

        splitter = "\n\n" if description else ""
        issues = "Issues: " + ", ".join(self.__commits.issues)

        return description + splitter + issues

    def __generate_description(self):
        return self.append_issues(
            "\n".join(list(map(lambda commit: commit.message.subject, self.__commits)))
        )

    def __merge_descriptions(self, existing_description: str, description: str) -> str:
        if not existing_description:
            return description

        splitter = SPLITTER.replace(
            "{{PULL_REQUEST_ID}}", str(self.__pull_request.pull_request_id)
        )
        splitter = splitter.replace("{{AUTHOR}}", self.__pull_request.author)
        return existing_description + splitter + description

    def update(self, title: str, description: str):
        self.__title = title
        self.__description = self.__remove_comments(description)

    @staticmethod
    def __remove_comments(description: str) -> str:
        return "\n".join(
            filter(lambda line: not line.startswith("#="), description.split("\n"))
        )



import pytest
from unittest.mock import patch, MagicMock

from pull_request_codecommit.git import Commit
from pull_request_codecommit.git.client import Client
from tests import COMMITS


def test_git_client() -> None:
    with pytest.raises(Exception) as exc:
        Client("fake-path")

    assert str(exc.value) == "The fake-path is not a directory."


@pytest.mark.parametrize(
    "path, remote",
    [
        ("my-path", "codecommit::eu-west-1://my-profile@my-repository"),
        (
            "my-other-path",
            "codecommit::eu-central-1://my-other-profile@my-other-repository",
        ),
    ],
)
@patch("pull_request_codecommit.git.client.subprocess.run")
@patch("pull_request_codecommit.git.client.os.path.isdir")
def test_remote(
    mock_isdir: MagicMock, mock_run: MagicMock, path: str, remote: str
) -> None:
    mock_isdir.return_value = True

    def execute(parameters, cwd, stdout):
        assert path == cwd
        assert -1 == stdout
        mock_stdout = MagicMock()
        mock_stdout.stdout = bytes(f"\n\n{remote}\n\n", "utf-8")
        return mock_stdout

    mock_run.side_effect = execute

    client = Client(path)

    assert client.remote() == remote
    assert client.remote("origin") == remote


@pytest.mark.parametrize(
    "path, branch",
    [
        ("my-path", "main"),
        (
            "my-other-path",
            "master",
        ),
    ],
)
@patch("pull_request_codecommit.git.client.subprocess.run")
@patch("pull_request_codecommit.git.client.os.path.isdir")
def test_current_branch(
    mock_isdir: MagicMock, mock_run: MagicMock, path: str, branch: str
) -> None:
    mock_isdir.return_value = True

    def execute(parameters, cwd, stdout):
        assert path == cwd
        assert -1 == stdout
        mock_stdout = MagicMock()
        mock_stdout.stdout = bytes(f"\n\n{branch}\n\n", "utf-8")
        return mock_stdout

    mock_run.side_effect = execute
    client = Client(path)
    assert client.current_branch == branch


@pytest.mark.parametrize(
    "path, branch",
    [
        ("my-path", "main"),
        (
            "my-other-path",
            "master",
        ),
    ],
)
@patch("pull_request_codecommit.git.client.subprocess.run")
@patch("pull_request_codecommit.git.client.os.path.isdir")
def test_get_commit_messages(
    mock_isdir: MagicMock, mock_run: MagicMock, path: str, branch: str
) -> None:
    mock_isdir.return_value = True

    def execute(parameters, cwd, stdout):
        assert path == cwd
        assert -1 == stdout
        mock_stdout = MagicMock()
        mock_stdout.stdout = bytes(COMMITS, "utf-8")
        return mock_stdout

    mock_run.side_effect = execute
    client = Client(path)

    commits = client.get_commit_messages("main")

    first: Commit = next(commits)
    second: Commit = next(commits)

    assert first == commits.first
    assert first.message.subject == "feat: my first commit"
    assert second.message.subject == "feat: my second commit"


@patch("pull_request_codecommit.git.client.subprocess.run")
@patch("pull_request_codecommit.git.client.os.path.isdir")
def test_get_no_commit_messages(mock_isdir: MagicMock, mock_run: MagicMock) -> None:
    mock_isdir.return_value = True

    def execute(parameters, cwd, stdout):
        assert -1 == stdout
        mock_stdout = MagicMock()
        mock_stdout.stdout = bytes("", "utf-8")
        return mock_stdout

    mock_run.side_effect = execute
    client = Client("my-path")

    commits = client.get_commit_messages("main")
    assert commits.first is None


@patch("pull_request_codecommit.git.client.subprocess.run")
@patch("pull_request_codecommit.git.client.os.path.isdir")
def test_push(mock_isdir: MagicMock, mock_run: MagicMock) -> None:
    mock_isdir.return_value = True

    def execute(parameters, cwd, stdout):
        assert parameters == ["git", "push", "--set-upstream", "origin", "HEAD"]
        assert -1 == stdout
        mock_stdout = MagicMock()
        mock_stdout.stdout = bytes("", "utf-8")
        return mock_stdout

    mock_run.side_effect = execute
    Client("my-path").push()


@patch("pull_request_codecommit.git.client.subprocess.run")
@patch("pull_request_codecommit.git.client.os.path.isdir")
def test_pull(mock_isdir: MagicMock, mock_run: MagicMock) -> None:
    mock_isdir.return_value = True

    def execute(parameters, cwd, stdout):
        assert parameters == ["git", "pull"]
        assert -1 == stdout
        mock_stdout = MagicMock()
        mock_stdout.stdout = bytes("", "utf-8")
        return mock_stdout

    mock_run.side_effect = execute
    Client("my-path").pull()


@patch("pull_request_codecommit.git.client.subprocess.run")
@patch("pull_request_codecommit.git.client.os.path.isdir")
def test_delete_branch(mock_isdir: MagicMock, mock_run: MagicMock) -> None:
    mock_isdir.return_value = True

    def execute(parameters, cwd, stdout):
        assert parameters == ["git", "branch", "-d", "feat/my-feature"]
        assert -1 == stdout
        mock_stdout = MagicMock()
        mock_stdout.stdout = bytes("", "utf-8")
        return mock_stdout

    mock_run.side_effect = execute
    Client("my-path").delete_branch("feat/my-feature")


@patch("pull_request_codecommit.git.client.subprocess.run")
@patch("pull_request_codecommit.git.client.os.path.isdir")
def test_checkout(mock_isdir: MagicMock, mock_run: MagicMock) -> None:
    mock_isdir.return_value = True

    def execute(parameters, cwd, stdout):
        assert parameters == ["git", "checkout", "master"]
        assert -1 == stdout
        mock_stdout = MagicMock()
        mock_stdout.stdout = bytes("", "utf-8")
        return mock_stdout

    mock_run.side_effect = execute
    Client("my-path").checkout("master")



from __future__ import annotations

import json
from typing import List, Optional
import subprocess


class Client:
    """
    Understands aws operations

    We use the AWS CLI for these operations so that we can use tha MFA toke cache from the cli.
    """

    def __init__(self, profile: Optional[str], region: Optional[str]) -> None:
        self.__base_command: List[str] = []
        self.__profile: Optional[str] = profile
        self.__region: Optional[str] = region

    @property
    def base_command(self) -> List[str]:
        base_command = ["aws", "--output", "json"]

        if self.__profile:
            base_command.extend(["--profile", self.__profile])

        if self.__region:
            base_command.extend(["--region", self.__region])

        return base_command

    def __execute(self, parameters: List[str]) -> str:
        command = self.base_command
        command.extend(parameters)
        response = subprocess.run(command, stdout=subprocess.PIPE)

        return response.stdout.decode("utf-8").strip("\n")

    def get_open_pull_request(
        self, repository: str, source: str, destination: str
    ) -> dict:
        response = self.__execute(
            [
                "codecommit",
                "list-pull-requests",
                "--repository-name",
                repository,
                "--pull-request-status",
                "OPEN",
            ]
        )
        data = json.loads(response)
        open_pull_requests = data.get("pullRequestIds")

        for pull_request_id in open_pull_requests:
            pr = self.__get_pull_request(pull_request_id)
            target = pr.get("pullRequestTargets", [{}])[0]

            if (
                target.get("sourceReference") == f"refs/heads/{source}"
                and target.get("destinationReference") == f"refs/heads/{destination}"
            ):
                return pr

        return {}

    def __get_pull_request(self, pull_request_id: int) -> dict:
        response = self.__execute(
            [
                "codecommit",
                "get-pull-request",
                "--pull-request-id",
                str(pull_request_id),
            ]
        )
        data = json.loads(response)
        return data.get("pullRequest", {})

    def create_pull_request(
        self,
        title: str,
        description: str,
        repository: str,
        source: str,
        destination: str,
    ) -> dict:
        response = self.__execute(
            [
                "codecommit",
                "create-pull-request",
                "--title",
                title,
                "--description",
                description,
                "--targets",
                f"repositoryName={repository}, sourceReference={source}, destinationReference={destination}",
            ]
        )
        data = json.loads(response)

        return data.get("pullRequest")

    def update_pull_request(
        self,
        pull_request_id: int,
        description: str,
    ) -> dict:
        response = self.__execute(
            [
                "codecommit",
                "update-pull-request-description",
                "--pull-request-id",
                str(pull_request_id),
                "--description",
                description,
            ]
        )
        data = json.loads(response)

        return data.get("pullRequest")

    def merge_pull_request(
        self, repository: str, pull_request_id: int, branch: str
    ) -> dict:
        response = self.__execute(
            [
                "codecommit",
                "merge-pull-request-by-fast-forward",
                "--pull-request-id",
                str(pull_request_id),
                "--repository-name",
                repository,
            ]
        )
        data = json.loads(response)
        pull_request = data.get("pullRequest")

        if pull_request.get("pullRequestStatus", "") == "CLOSED":
            self.__execute(
                [
                    "codecommit",
                    "delete-branch",
                    "--repository-name",
                    repository,
                    "--branch-name",
                    branch,
                ]
            )

        return pull_request
